"""
NestedHashOptional.py

Implements a depending optional operator.

Autor: Maribel Acosta
Autor: Gabriela Montoya
Date: November 18th, 2013
"""

from .OperatorStructures import Table, Partition, Record
from multiprocessing import Queue, Process
from time import time
import string, sys
from queue import Empty
from DeTrusty.Operators.Optional import Optional


class NestedHashOptional(Optional):

    def __init__(self, vars_left, vars_right):
        self.left_table = dict()
        self.right_table = dict()
        self.qresults = Queue()
        self.bag = []
        self.vars_left = set(vars_left)
        self.vars_right = set(vars_right)
        self.vars = list(self.vars_left & self.vars_right)

    def instantiate(self, d):
        newvars_left = self.vars_left - set(d.keys())
        newvars_right = self.vars_right - set(d.keys())
        return NestedHashOptional(newvars_left, newvars_right)

    def execute(self, left_queue, right_operator, out, processqueue=Queue()):
        self.left_queue = left_queue
        self.right_operator = right_operator
        self.qresults = out

        tuple1 = None
        tuple2 = None

        right_queues = dict()

        while (not (tuple1 == "EOF") or (len(right_queues) > 0)):

            # Try to get and process tuple from left queue
            if not (tuple1 == "EOF"):
                try:
                    tuple1 = self.left_queue.get(False)
                    if not (tuple1 == "EOF"):
                        self.bag.append(tuple1)
                    # print "tuple1: "+str(tuple1)
                    instance = self.probeAndInsert1(tuple1, self.right_table,
                                                    self.left_table, time())
                    # print "instance:", instance
                    if instance:  # the join variables have not been used to
                        # instanciate the right_operator

                        new_right_operator = self.makeInstantiation(tuple1,
                                                                    self.right_operator)
                        # print "new op: "+str(new_right_operator)
                        resource = self.getResource(tuple1)
                        queue = Queue()
                        right_queues[resource] = queue
                        # print "new_right_operator.__class__", new_right_operator.__class__
                        # print "new_right_operator.left.__class__", new_right_operator.left.__class__
                        new_right_operator.execute(queue)
                        # p2 = Process(target=new_right_operator.execute, args=(queue,))
                        # p2.start()
                except Empty:
                    pass
                except TypeError:
                    # TypeError: in resource = resource + tuple[var], when the tuple is "EOF".
                    pass
                except:
                    # print "Unexpected error:", sys.exc_info()[0]
                    # raise
                    pass

            toRemove = []  # stores the queues that have already received all its tuples

            for r in right_queues:
                try:
                    q = right_queues[r]
                    tuple2 = q.get(False)
                    if (tuple2 == "EOF"):
                        toRemove.append(r)
                    else:
                        self.probeAndInsert2(r, tuple2, self.left_table,
                                             self.right_table, time())
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    # TypeError: in att = att + tuple[var], when the tuple is "EOF".
                    # print "Unexpected error:", sys.exc_info()[0]
                    pass

            for r in toRemove:
                del right_queues[r]

        # This is the optional: Produce tuples that haven't matched already.
        for tuple in self.bag:
            res_right = {}
            for var in self.vars_right:
                res_right.update({var: ''})
            res = res_right
            res.update(tuple)
            self.qresults.put(res)

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        # return

    def getResource(self, tuple):
        resource = ''
        for var in self.vars:
            resource = resource + tuple[var]
        return resource

    def makeInstantiation(self, tuple, operator):
        # print "making instantiation", tuple, operator
        d = {}
        for var in self.vars:
            v = tuple[var]
            if string.find(v, "http") == 0:  # uris must be passed between < .. >
                v = "<" + v + ">"
            else:
                v = '"' + v + '"'
            d[var] = v
        # print "d", d
        # print 'type(operator) in makeInstantiation in Nested Optiional', type(operator)
        new_operator = operator.instantiate(d)
        return new_operator

    def probeAndInsert1(self, tuple, table1, table2, time):
        # print "probeAndInsert1", tuple
        record = Record(tuple, time, 0)
        r = self.getResource(tuple)
        if r in table1:
            records = table1[r]
            for t in records:
                if t.ats > record.ats:
                    continue
                x = t.tuple.copy()
                x.update(tuple)
                self.qresults.put(x)
                # Delete tuple from bag.
                try:
                    self.bag.remove(tuple)
                except ValueError:
                    pass
        p = table2.get(r, [])
        i = (p == [])
        p.append(record)
        table2[r] = p
        return i

    def probeAndInsert2(self, resource, tuple, table1, table2, time):
        # print "probeAndInsert2", tuple
        record = Record(tuple, time, 0)
        if resource in table1:
            records = table1[resource]
            for t in records:
                if t.ats > record.ats:
                    continue
                x = t.tuple.copy()
                x.update(tuple)
                self.qresults.put(x)

                # Delete tuple from bag.
                try:
                    self.bag.remove(t.tuple)
                except ValueError:
                    pass

        p = table2.get(resource, [])
        p.append(record)
        table2[resource] = p



"""
Created on Jul 10, 2011

Implements the Xgjoin operator.
The intermediate results are represented as a queue.

@author: Maribel Acosta Deibe
"""

import signal
from multiprocessing import Queue
from queue import Empty
from time import time
from tempfile import NamedTemporaryFile
from os import remove
from DeTrusty.Operators.Join import Join
from .OperatorStructures import Record, RJTTail, FileDescriptor


class Xgjoin(Join):

    def __init__(self, vars):
        self.left_table  = dict()
        self.right_table = dict()
        self.qresults    = Queue()
        self.vars        = vars

        # Second stage settings
        self.secondStagesTS     = []
        self.lastSecondStageTS  = float("-inf")
        self.timeoutSecondStage = 100000000
        self.sourcesBlocked     = False

        # Main memory settings
        self.memorySize   = 100000000        # Represents the main memory size (# tuples).OLD:Represents the main memory size (in KB).
        self.fileDescriptor_left = {}
        self.fileDescriptor_right = {}
        self.memory_left  = 0
        self.memory_right = 0

        self.leftcount = 0
        self.rightcount = 0

    def instantiate(self, d):
        newvars = self.vars - set(d.keys())
        return Xgjoin(newvars)

    def instantiateFilter(self, instantiated_vars, filter_str):
        newvars = self.vars - set(instantiated_vars)
        return Xgjoin(newvars)

    def execute(self, left, right, out, processqueue=Queue()):
        # Executes the Xgjoin.
        self.left     = left
        self.right    = right
        self.qresults = out

        # Initialize tuples.
        tuple1 = None
        tuple2 = None

        # Create alarm to go to stage 2.
        signal.signal(signal.SIGALRM, self.stage2)

        # Get the tuples from the queues.
        while tuple1 != "EOF" or tuple2 != "EOF":

            # Try to get and process tuple from left queue.
            if tuple1 != "EOF":
                try:
                    tuple1 = self.left.get(False)
                    # print "tuple1", tuple1
                    self.leftcount += 1
                    signal.alarm(self.timeoutSecondStage)
                    self.stage1(tuple1, self.left_table, self.right_table)
                    self.memory_right += 1
                except Empty:
                    # Empty: in tuple1 = self.left.get(False), when the queue is empty.
                    pass
                except TypeError as te:
                    print("TypeError: in resource = resource + tuple[var]", tuple, te)
                    # TypeError: in resource = resource + tuple[var], when the tuple is "EOF".
                    pass
                except IOError:
                    # IOError: when a tuple is received, but the alarm is fired.
                    self.sourcesBlocked = False
                    pass

            # Try to get and process tuple from right queue.
            if tuple2 != "EOF":
                try:
                    tuple2 = self.right.get(False)
                    # print "tuple2", tuple2
                    self.rightcount +=1
                    signal.alarm(self.timeoutSecondStage)
                    self.stage1(tuple2, self.right_table, self.left_table)
                    self.memory_left += 1
                except Empty:
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    pass
                except TypeError as te:
                    print("TypeError: in resource = resource + tuple[var]", tuple, te)
                    # TypeError: in resource = resource + tuple[var], when the tuple is "EOF".
                    pass
                except IOError:
                    # IOError: when a tuple is received, but the alarm is fired.
                    self.sourcesBlocked = False
                    pass

            #print "(LEFT, RIGHT) = >", self.leftcount, self.rightcount, self.vars
            if (len(self.left_table) + len(self.right_table) >= self.memorySize):
                self.flushRJT()
                #print "Flushed RJT!"

        # Turn off alarm to stage 2.
        signal.alarm(0)
        # Perform the last probes.
        self.stage3()
        return

    def stage1(self, tuple, tuple_rjttable, other_rjttable):
        #print " Stage 1: While one of the sources is sending data."
        if (tuple != "EOF"):
            # Get the resource associated to the tuples.
            resource = ''
            #print(tuple)
            for var in self.vars:
                if var in tuple:
                    val = tuple[var]
                    if "^^<" in val:
                        val = val[:val.find('^^<')]
                    resource = resource + str(val)

            # Probe the tuple against its RJT table.
            probeTS = self.probe(tuple, resource, tuple_rjttable)

            # Create the records.
            record = Record(tuple, probeTS, time(), float("inf"))

            # Insert the record in the other RJT table.
            if resource in other_rjttable:
                other_rjttable.get(resource).updateRecords(record)
                other_rjttable.get(resource).setRJTProbeTS(probeTS)
                #other_rjttable.get(resource).append(record)
            else:
                tail = RJTTail(record, probeTS)
                other_rjttable[resource] = tail
                #other_rjttable[resource] = [record]

    def stage2(self, signum, frame):
        #print " Stage 2: When both sources become blocked."
        self.sourcesBlocked = True

        # Get common resources.
        resources1 = set(self.left_table.keys()) & set(self.fileDescriptor_right.keys())
        resources2 = set(self.right_table.keys()) & set(self.fileDescriptor_left.keys())

        # Iterate while there are common resources and both sources are blocked.
        while((resources1 or resources2) and self.sourcesBlocked):

            if (resources1):
                resource = resources1.pop()
                rjts1 = self.left_table[resource].records
                for rjt1 in rjts1:
                    probed = self.probeFile(rjt1, self.fileDescriptor_right, resource, 2)
                    if (probed):
                        rjt1.probeTS = time()

            elif (resources2):
                resource = resources2.pop()
                rjts1 = self.right_table[resource].records
                for rjt1 in rjts1:
                    probed = self.probeFile(rjt1, self.fileDescriptor_left, resource, 2)
                    if (probed):
                        rjt1.probeTS = time()

        # End of second stage.
        self.lastSecondStageTS = time()
        self.secondStagesTS.append(self.lastSecondStageTS)

#        fd_left  = len(set(map(FileDescriptor.getSize, self.fileDescriptor_left.values())))
#        fd_right = len(set(map(FileDescriptor.getSize, self.fileDescriptor_right.values())))
#        count = 0
#
#        while ((count < fd_left + fd_right) and self.sourcesBlocked):
#
#            (largestRJTs, table) = self.getLargestRJTs(count)
#            #print "Largests RJT:", largestRJTs
#            common_resources = set(largestRJTs.keys()) & set(table.keys())
#            print "Common R:", common_resources
#            for resource in common_resources:
#                rjts1 = table[resource].records
#                for rjt1 in rjts1:
#                    self.probeFile(rjt1, largestRJTs, resource, 2)
#
#            count = count + 1
#
#        self.lastSecondStageTS = time()
#        self.secondStagesTS.append(self.lastSecondStageTS)
#        print "----------------END Second Stage!"

    def stage3(self):
        #print "Stage 3: When both sources sent all the data."

        # RJTs in main (left) memory are probed against RJTs in secondary (right) memory.
        common_resources = set(self.left_table.keys()) & set(self.fileDescriptor_right.keys())
        for resource in common_resources:
            rjts1 = self.left_table[resource].records
            for rjt1 in rjts1:
                self.probeFile(rjt1, self.fileDescriptor_right, resource, 3)

        # RJTs in main (right) memory are probed against RJTs in secondary (left) memory.
        common_resources = set(self.right_table.keys()) & set(self.fileDescriptor_left.keys())
        for resource in common_resources:
            rjts1 = self.right_table[resource].records
            for rjt1 in rjts1:
                self.probeFile(rjt1, self.fileDescriptor_left, resource, 3)

        # RJTs in secondary memory are probed to produce new results.
        common_resources = set(self.fileDescriptor_left.keys()) & set(self.fileDescriptor_right.keys())
        for resource in common_resources:
            file1 = open(self.fileDescriptor_right[resource].file.name)
            rjts1 = file1.readlines()
            for rjt1 in rjts1:
                (tuple1, probeTS1, insertTS1, flushTS1) = rjt1.split('|')
                self.probeFile(Record(eval(tuple1), float(probeTS1), float(insertTS1), float(flushTS1)), self.fileDescriptor_left, resource, 3)
            file1.close()

        for resource in common_resources:
            file1 = open(self.fileDescriptor_left[resource].file.name)
            rjts1 = file1.readlines()
            for rjt1 in rjts1:
                (tuple1, probeTS1, insertTS1, flushTS1) = rjt1.split('|')
                self.probeFile(Record(eval(tuple1), float(probeTS1), float(insertTS1), float(flushTS1)), self.fileDescriptor_right, resource, 3)
            file1.close()

        # Delete files from secondary memory.
        for resource in self.fileDescriptor_left:
            remove(self.fileDescriptor_left[resource].file.name)

        for resource in self.fileDescriptor_right:
            remove(self.fileDescriptor_right[resource].file.name)

        # Put EOF in queue and exit.
        self.qresults.put("EOF")

    def probe(self, tuple, resource, rjttable):
        # Probe a tuple against its corresponding table.

        probeTS = time()
        # If the resource is in table, produce results.
        if resource in rjttable:
            rjttable.get(resource).setRJTProbeTS(probeTS)
            list_records = rjttable[resource].records

            for record in list_records:
                res = {}
                res.update(record.tuple)
                #res = record.tuple.copy()
                res.update(tuple)
                self.qresults.put(res)
                #print hex(id(self)), "res:", res

        return probeTS

    def probeFile(self, rjt1, filedescriptor2, resource, stage):
        # Probe an RJT against its corresponding partition in secondary memory.

        file2 = open(filedescriptor2[resource].file.name, 'r')
        rjts2 = file2.readlines()
        st = ""
        probed = False

        for rjt2 in rjts2:
            (tuple2, probeTS2, insertTS2, flushTS2) = rjt2.split('|')
            probedStage1 = False
            probedStage2 = False

            #Checking Property 2: Probed in stage 2.
            for ss in self.secondStagesTS:
                if (float(flushTS2) < rjt1.insertTS and rjt1.insertTS < ss and  ss < rjt1.flushTS):
                    probedStage2 = True
                    break

            # Checking Property 1: Probed in stage 1.
            if (rjt1.probeTS < float(flushTS2)):
                probedStage1 = True

            # Produce result if it has not been produced.
            if (not(probedStage1) and not(probedStage2)):
                res = rjt1.tuple.copy()
                res.update(eval(tuple2))
                self.qresults.put(res)
                probed = True

            # Update probeTS of tuple2.
            stprobeTS  = "%.40r" % (time())
            st = st + tuple2 + '|' + stprobeTS + '|' + insertTS2 + '|' + flushTS2

        file2.close()

        # Update file2 if in stage 2.
        if ((stage == 2) and probed):
            file2 = open(filedescriptor2[resource].file.name, 'w')
            file2.write(st)
            file2.close()

        return probed

    def flushRJT(self):
        # Flush an RJT to secondary memory.

        # Choose a victim from each partition (table).
        (resource_to_flush1, tail_to_flush1, least_ts1) = self.getVictim(self.left_table)
        (resource_to_flush2, tail_to_flush2, least_ts2) = self.getVictim(self.right_table)

        # Flush resource from left table.
        if least_ts1 <= least_ts2:
            file_descriptor = self.fileDescriptor_left
            table = self.left_table
            resource_to_flush = resource_to_flush1
            tail_to_flush = tail_to_flush1

        # Flush resource from right table.
        if least_ts2 < least_ts1:
            file_descriptor = self.fileDescriptor_right
            table = self.right_table
            resource_to_flush = resource_to_flush2
            tail_to_flush = tail_to_flush2

        # Create flush timestamp.
        flushTS = time()

        # Update file descriptor
        if (resource_to_flush in file_descriptor):
            lentail = file_descriptor[resource_to_flush].size
            file = open(file_descriptor[resource_to_flush].file.name, 'a')
            file_descriptor.update({resource_to_flush: FileDescriptor(file, len(tail_to_flush.records) + lentail, flushTS)})
        else:
            file = NamedTemporaryFile(suffix=".rjt", prefix="", delete=False)
            file_descriptor.update({resource_to_flush: FileDescriptor(file, len(tail_to_flush.records), flushTS)})

        # Flush tail in file.
        for record in tail_to_flush.records:
            sttuple    = str(record.tuple)
            stprobeTS  = "%.40r" % (record.probeTS)
            stinsertTS = "%.40r" % (record.insertTS)
            stflushTS  = "%.40r" % (flushTS)

            file.write(sttuple + '|')
            file.write(stprobeTS + '|')
            file.write(stinsertTS + '|')
            file.write(stflushTS + '\n')
        file.close()

        # Delete resource from main memory.
        del table[resource_to_flush]

    def getVictim(self, table):
        # Selects a victim from a partition in main memory to flush.

        resource_to_flush = ""
        tail_to_flush = RJTTail([], 0)
        least_ts = float("inf")

        for resource, tail in table.items():
            resource_ts = tail.rjtProbeTS
            if ((resource_ts < least_ts) or
                (resource_ts == least_ts and len(tail.records) > len(tail_to_flush.records))):
                resource_to_flush = resource
                tail_to_flush = tail
                least_ts = resource_ts

        #print "Victim chosen:", resource_to_flush, "TS:", least_ts, "LEN:", len(tail_to_flush.records)
        return (resource_to_flush, tail_to_flush, least_ts)

    def getLargestRJTs(self, i):
        # Selects the i-th largest RJT stored in secondary memory.

        sizes1 = set(map(FileDescriptor.getSize, self.fileDescriptor_left.values()))
        sizes2 = set(map(FileDescriptor.getSize, self.fileDescriptor_right.values()))

        sizes1 = list(sizes1)
        sizes2 = list(sizes2)

        sizes1.sort()
        sizes2.sort()

        if (sizes1 and sizes2):
            if (sizes1[len(sizes1)-1] > sizes2[len(sizes2)-1]):
                file_descriptor = self.fileDescriptor_left
                max_len = sizes1[len(sizes1)-(i+1)]
                table = self.right_table
            else:
                file_descriptor = self.fileDescriptor_right
                max_len = sizes2[len(sizes2)-(i+1)]
                table = self.left_table
        elif (sizes1):
            file_descriptor = self.fileDescriptor_left
            max_len = sizes1[len(sizes1)-(i+1)]
            table = self.right_table
        else:
            file_descriptor =self.fileDescriptor_right
            max_len = sizes2[len(sizes2)-(i+1)]
            table = self.left_table

        largestRJTs = {}

        for resource, fd in file_descriptor.items():
            if (fd.size == max_len):
                largestRJTs[resource] = fd

        return (largestRJTs, table)



import urllib.parse
from ply import lex, yacc
# not all module needed (TODO: check after integrating expression to Having, Aggregate)
from .services import Query, Argument, Triple, UnionBlock, JoinBlock, Optional, Filter, Expression, Values, Bind, Aggregate, Having, HavingHelper, Service


# Lexer
reserved = {
    'PREFIX': 'PREFIX',
    'UNION': 'UNION',
    'FILTER': 'FILTER',
    'OPTIONAL': 'OPTIONAL',
    'VALUES' : 'VALUES',
    'UNDEF' : 'UNDEF',
    'BIND': 'BIND',
    'AS': 'AS',
    'SERVICE': 'SERVICE',
    'SELECT': 'SELECT',
    'DISTINCT': 'DISTINCT',
    'WHERE': 'WHERE',
    'LIMIT': 'LIMIT',
    'OFFSET': 'OFFSET',
    'ORDER': 'ORDER',
    'BY': 'BY',
    'DESC': 'DESC',
    'ASC': 'ASC',
    'IN' : 'IN',
    'NOT' : 'NOT',
    'FALSE': 'FALSE',
    'TRUE': 'TRUE',
    'GROUP': 'GROUP',
    'COUNT': 'COUNT',
    'SUM': 'SUM',
    'AVG': 'AVG',
    'MIN': 'MIN',
    'MAX': 'MAX',
    'SAMPLE': 'SAMPLE',
    'GROUP_CONCAT': 'GROUP_CONCAT',
    'SEPARATOR': 'SEPARATOR',
    'HAVING': 'HAVING',
    'BOUND': 'BOUND',
    'REGEX': 'REGEX',
    'ISIRI': 'ISIRI',
    'ISURI': 'ISURI',
    'ISBLANK': 'ISBLANK',
    'ISLITERAL': 'ISLITERAL',
    'IRI': 'IRI',
    'ABS': 'ABS',
    'CEIL': 'CEIL',
    'LANG': 'LANG',
    'ROUND': 'ROUND',
    'STRLEN': 'STRLEN',
    'FLOOR': 'FLOOR',
    'DATATYPE': 'DATATYPE',
    'SAMETERM': 'SAMETERM',
    'LANGMATCHES': 'LANGMATCHES',
    'STR': 'STR',
    'UCASE': 'UCASE',
    'LCASE': 'LCASE',
    'ENCODE_FOR_URI': 'ENCODE_FOR_URI',
    'YEAR': 'YEAR',
    'MONTH': 'MONTH',
    'DAY': 'DAY',
    'HOURS': 'HOURS',
    'MINUTES': 'MINUTES',
    'SECONDS': 'SECONDS',
    'CONTAINS': 'CONTAINS',
    'STRSTARTS': 'STRSTARTS',
    'STRENDS': 'STRENDS',
    'STRBEFORE': 'STRBEFORE',
    'STRAFTER': 'STRAFTER',
    'STRLANG': 'STRLANG',
    'STRDT': 'STRDT',
    'SUBSTR': 'SUBSTR',
    'TIMEZONE': 'TIMEZONE',
    'TZ': 'TZ',
    'CONCAT': 'CONCAT',
    'REPLACE': 'REPLACE',
    'EXIST': 'EXIST',
    'MD5': 'MD5',
    'COALESCE': 'COALESCE',
    'IF': 'IF',
    'SHA1': 'SHA1',
    'SHA256': 'SHA256',
    'SHA384': 'SHA384',
    'SHA512': 'SHA512',
    'ISNUMERIC': 'ISNUMERIC',
    'RAND': 'RAND',
    'NOW': 'NOW',
    'UUID': 'UUID',
    'STRUUID': 'STRUUID',
    'BNODE': 'BNODE'
}


tokens = [
    # "RDFTYPE",
    "CONSTANT",
    "URI",
    "NUMBER",
    "VARIABLE",
    "LKEY",
    "RKEY",
    "COLON",
    "SEMICOLON",
    "POINT",
    "COMA",
    "LPAR",
    "RPAR",
    "ID",
    "NIL",
    "EQUALS",
    "NEQUALS",
    "LESS",
    "LESSEQ",
    "GREATER",
    "GREATEREQ",
    "NEG",
    "AND",
    "OR",
    "ALL",
    "PLUS",
    "MINUS",
    "DIV",
    "DOUBLE",
    "INTEGER",
    "DECIMAL",
    "FLOAT",
    "STRING",
    "BOOLEAN",
    "DATETIME",
    "NONPOSINT",
    "NEGATIVEINT",
    "LONG",
    "INT",
    "SHORT",
    "BYTE",
    "NONNEGINT",
    "UNSIGNEDLONG",
    "UNSIGNEDINT",
    "UNSIGNEDSHORT",
    "UNSIGNEDBYTE",
    "POSITIVEINT"
    ] + list(reserved.values())

# t_RDFTYPE = r"a"


def t_ID(t):
    r'[a-zA-Z_][a-zA-Z_0-9\-]*'
    # print t
    t.type = reserved.get(t.value.upper(), 'ID')    # Check for reserved words
    return t


t_CONSTANT = r"(\"|\')[^\"\'\n\r]*(\"|\')((@[a-z][a-z]) | (\^\^[<](https?|ftp|file)://[-a-zA-Z0-9+&@#/%?=~_|!:,.;]*[-a-zA-Z0-9+&@#/%=~_|][>]))?" # [<]http[://]+www[.]w3[.]org[/]2001[/]XMLSchema[#]\w+
t_URI = r"<\S+>"
t_NUMBER = r"([0-9])+"
t_VARIABLE = r"([\?]|[\$])([A-Z]|[a-z])\w*"
t_LKEY = r"\{"
t_LPAR = r"\("
t_RPAR = r"\)"
t_COLON = r"\:"
t_SEMICOLON = r";"
# t_RDFTYPE = r"a"
# t_RKEY = r"(\.)?\}"
t_RKEY = r"(\.)?\s*\}"
t_POINT = r"\."
t_COMA = r"\,"

t_EQUALS = r"="
t_NEQUALS = r"\!="
t_LESS = r"<"
t_LESSEQ = r"<="
t_GREATER = r">"
t_GREATEREQ = r">="
t_NEG  =  r"\!"
t_AND = r"\&\&"
t_OR = r"\|\|"
# t_NIL =

t_ALL = r"[*]"
t_PLUS = r"\+"
t_MINUS = r"\-"
t_DIV = r"/"

t_DOUBLE = r"xsd\:double"
t_INTEGER = r"xsd\:integer"
t_DECIMAL = r"xsd\:decimal"
t_FLOAT = r"xsd\:float"
t_STRING = r"xsd\:string"
t_BOOLEAN = r"xsd\:boolean"
t_DATETIME = r"xsd\:dateTime"
t_NONPOSINT = r"xsd\:nonPositiveInteger"
t_NEGATIVEINT = r"xsd\:negativeInteger"
t_LONG = r"xsd\:long"
t_INT = r"xsd\:int"
t_SHORT = r"xsd\:short"
t_BYTE = r"xsd\:byte"
t_NONNEGINT = r"xsd\:nonNegativeInteger"
t_UNSIGNEDLONG = r"xsd\:unsignedLong"
t_UNSIGNEDINT = r"xsd\:unsignedInt"
t_UNSIGNEDSHORT = r"xsd\:unsignedShort"
t_UNSIGNEDBYTE = r"xsd\:unsignedByte"
t_POSITIVEINT = r"xsd\:positiveInteger"

t_ignore = ' \t\n'

xsd = "http://www.w3.org/2001/XMLSchema#"


def t_error(t):
    raise TypeError("Unknown text '%s' in line %d " % (t.value, t.lexer.lineno, ))


# Define a rule so we can track line numbers
def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)


lexer = lex.lex()


# Parser


# TODO: merge the services.py 

def p_parse_sparql(p):
    """
    parse_sparql : prefix_list query group_by having_clause order_by limit offset values_clause
    """
    (vs, ts, d) = p[2]
    p[0] = Query(prefs=p[1], args=vs, body=ts, distinct=d, values=p[8], group_by=p[3], order_by=p[5], limit=p[6], offset=p[7], having=p[4])


def p_empty(p):
    """
    empty :
    """
    pass


# in Specification PREFIX : <http://example.com/> (without ID) is a valid declaration, here not

def p_prefix_list(p):
    """
    prefix_list : prefix prefix_list
    """
    p[0] = [p[1]] + p[2]


def p_empty_prefix_list(p):
    """
    prefix_list : empty
    """
    p[0] = []


def p_prefix(p):
    """
    prefix : PREFIX uri
    """
    p[0] = p[2]


def p_uri_0(p):
    """
    uri : ID COLON ID
    """
    p[0] = p[1] + p[2] + p[3]

def p_uri_1(p):
    """
    uri : ID COLON URI
    """
    p[0] = p[1] + p[2] + p[3]


def p_uri_2(p):
    """
    uri : URI
    """
    p[0] = p[1]


def p_uri_3(p):
    """
    uri : ID COLON YEAR
    """
    p[0] = p[1] + p[2] + p[3]

def p_uri_4(p):
    """
    uri : ID COLON ID NUMBER
    """
    p[0] = p[1] + p[2] + p[3] + p[4]

def p_uri_5(p):
    """
    uri : ID COLON NUMBER
    """
    p[0] = p[1] + p[2] + p[3]

########################################################################

# REDUCED not yet implemented, necessary?

def p_query_0(p):
    """
    query : SELECT distinct var_list WHERE LKEY group_graph_pattern RKEY
    """
    p[0] = (p[3], p[6], p[2])


def p_query_1(p):
    """
    query : SELECT distinct ALL WHERE LKEY group_graph_pattern RKEY
    """
    p[0] = ([], p[6], p[2])


def p_distinct_0(p):
    """
    distinct : DISTINCT
    """
    p[0] = True


def p_distinct_1(p):
    """
    distinct : empty
    """
    p[0] = False


def p_single_var_list_0(p): 
    """
    var_list : VARIABLE
    """
    p[0] = [Argument(p[1], False)]


def p_single_var_list_1(p): 
    """
    var_list : LPAR expression AS VARIABLE RPAR
    """
    p[2].alias = p[4]
    p[0] = [p[2]]


# def p_single_var_list_2(p): 
    # """
    # var_list : VARIABLE AS VARIABLE
    # """
    # p[0] = [Argument(p[1], False, alias=p[3])]


def p_var_list_0(p):
    """
    var_list : VARIABLE var_list 
    """
    p[0] = [Argument(p[1], False)] + p[2]


def p_var_list_1(p):
    """
    var_list : LPAR expression AS VARIABLE RPAR var_list 
    """
    p[2].alias = p[4]
    p[0] = [p[2]] + p[6]


# def p_var_list_2(p):
    # """
    # var_list : VARIABLE AS VARIABLE var_list
    # """
    # p[0] = [Argument(p[1], False, alias=p[3])] + p[4]


################################################################
# TODO: Adjust until Xgroupby.py                               #
################################################################


def p_group_by_0(p):
   """
   group_by : GROUP BY group_var group_var_list
   """
   p[0] = [p[3]] + p[4]


def p_group_by_1(p):
   """
   group_by : empty
   """
   p[0] = []


def p_group_var_0(p):                          
   """                                        
   group_var : VARIABLE                        
   """
   p[0] = Argument(p[1], False)


def p_group_var_1(p):                          
   """                                        
   group_var : LPAR expression AS VARIABLE RPAR                         
   """
   p[0] = Argument(p[2], False, alias=p[4])


def p_group_var_2(p):                          
   """                                        
   group_var : LPAR expression RPAR                         
   """
   p[0] = Argument(p[1], False)


def p_group_var_3(p):                          
   """                                        
   group_var : built_in_call
   """
   p[0] = Argument(p[1], False)


def p_group_var_list_0(p):
   """
   group_var_list : group_var group_var_list
   """
   p[0] = [p[1]] + p[2]


def p_group_var_list_1(p):
   """
   group_var_list : empty
   """
   p[0] = []



########################### TODO #############################################
# -> Adjust the orderby module & services.py (for multiple condition)        #
##############################################################################


def p_order_by_0(p):
    """
    order_by : ORDER BY order_by_condition
    """
    p[0] = p[3]


def p_order_by_1(p):
    """
    order_by : empty
    """
    p[0] = []


def p_order_by_condition_0(p):
    """
    order_by_condition : ASC constraint order_by_condition 
    """
    p[0] = [Argument(p[2], desc=False)] + p[3]


def p_order_by_condition_1(p):
    """
    order_by_condition : DESC constraint order_by_condition
    """
    p[0] = [Argument(p[2], desc=True)] + p[3]
    

def p_order_by_condition_2(p):
    """
    order_by_condition : constraint order_by_condition
    """
    p[0] =  [Argument(p[1], desc=False)] + p[2] # by default ASC


def p_order_by_condition_3(p):
    """
    order_by_condition : ASC VARIABLE order_by_condition
    """
    p[0] =  [Argument(p[2], desc=False)] + p[3]


def p_order_by_condition_4(p):
    """
    order_by_condition : DESC VARIABLE order_by_condition
    """
    p[0] =  [Argument(p[2], desc=True)] + p[3]


def p_order_by_condition_5(p):
    """
    order_by_condition : VARIABLE 
    """
    p[0] =  [Argument(p[1], desc=True)]


############################################################
# -> Constraint does not include FunctionCall from spec.   #
############################################################


def p_constraint_0(p):
    """
    constraint : bracketted_expression 
    """
    p[0] = p[1]


def p_constraint_1(p):
    """
    constraint : built_in_call 
    """
    p[0] = p[1]


############################################################


def p_limit_0(p):
    """
    limit : LIMIT NUMBER
    """
    p[0] = p[2]


def p_limit_1(p):
    """
    limit : empty
    """
    p[0] = -1


def p_offset_0(p):
    """
    offset : OFFSET NUMBER
    """
    p[0] = p[2]


def p_offset_1(p):
    """
    offset : empty
    """
    p[0] = -1


######################### TODO #############################
# -> Multiple constrain ?                                  #
# -> fix until Xhaving.py                                  #
############################################################


def p_having_clause_0(p):
    """
    having_clause : empty
    """
    p[0] = None


def p_having_clause_1(p):
    """
    having_clause : HAVING constraint
    """
    p[0] = p[1]


############################################################


def p_values_clause_0(p):
    """
    values_clause : VALUES LPAR var_list RPAR LKEY data_block RKEY
    """
    p[0] = Values(p[3], p[6])


def p_values_clause_1(p):
    """
    values_clause : VALUES var_list LKEY data_block RKEY
    """
    p[0] = Values(p[2], p[4])


def p_values_clause_2(p):
    """
    values_clause : empty
    """
    p[0] = None 


def p_data_block_0(p):
    """
    data_block : data_block LPAR values_list RPAR
    """ 
    p[0] = p[1] + [p[3]]


def p_data_block_1(p):
    """
    data_block : data_block values_list
    """ 
    p[0] = p[1] + [p[2]]


def p_data_block_2(p):
    """
    data_block : values_list
    """
    p[0] = [p[1]]


def p_data_block_3(p):
    """
    data_block : LPAR values_list RPAR
    """
    p[0] = [p[2]]


def p_values_list_0(p):
    """
    values_list : values_list value_arg
    """
    p[0] = p[1] + [p[2]]


def p_values_list_1(p):
    """
    values_list : value_arg
    """
    p[0] = [p[1]]


def p_value_arg_0(p):
    """
    value_arg : uri
    """
    p[0] = Argument(p[1], True)


def p_value_arg_1(p):
    """
    value_arg : rdf_literal
    """
    p[0] = p[1]


def p_value_arg_2(p):
    """
    value_arg : numeric_literal
    """
    p[0] = p[1]


def p_value_arg_3(p):
    """
    value_arg : boolean_literal
    """
    p[0] = p[1]


def p_value_arg_4(p):
    """
    value_arg : UNDEF
    """
    p[0] = None 

############################################################

# section to be re-checked at the end

def p_ggp_0(p):
    """
    group_graph_pattern : union_block
    """
    p[0] = UnionBlock(p[1])


def p_service(p):
    """
    service : SERVICE uri LKEY group_graph_pattern_service RKEY
    """
    p[0] = Service(p[2], p[4])


def p_ggp_service_0(p):
    """
    group_graph_pattern_service : union_block_service
    """
    p[0] = UnionBlock(p[1])


def p_union_block_0(p):
    """
    union_block : pjoin_block rest_union_block POINT pjoin_block
    """
    punion = [JoinBlock(p[1])] + p[2]
    pjoin = [UnionBlock(punion)] + p[4]
    p[0] = [JoinBlock(pjoin)]


def p_union_block_1(p):
    """
    union_block : pjoin_block rest_union_block pjoin_block
    """
    punion = [JoinBlock(p[1])] + p[2]
    if p[3] != []:
        pjoin = [UnionBlock(punion)] + p[3]
        p[0] = [JoinBlock(pjoin)]
    else:
        p[0] = [JoinBlock(p[1])] + p[2]


def p_union_block_2(p):
    """
    union_block : pjoin_block rest_union_block
    """
    p[0] = [JoinBlock(p[1])] + p[2]


def p_union_block_service_0(p):
    """
    union_block_service : join_block_service rest_union_block_service
    """
    p[0] = [JoinBlock(p[1])] + p[2]


def p_rest_union_block_service_0(p):
    """
    rest_union_block_service : empty
    """
    p[0] = []


def p_rest_union_block_service_1(p):
    """
    rest_union_block_service : UNION LKEY join_block_service rest_union_block_service RKEY rest_union_block_service
    """
    p[0] = [JoinBlock(p[3])] + p[4] + p[6]


def p_ppjoin_block_0(p):
    """
    pjoin_block : LKEY join_block RKEY
    """
    p[0] = p[2]


def p_ppjoin_block_1(p):
    """
    pjoin_block : join_block
    """
    p[0] = p[1]


def p_ppjoin_block_2(p):
    """
    pjoin_block : empty
    """
    p[0] = []


def p_rest_union_block_0(p):
    """
    rest_union_block : empty
    """
    p[0] = []


def p_rest_union_block_1(p):
    """
    rest_union_block : UNION LKEY join_block rest_union_block RKEY rest_union_block
    """
    p[0] = [JoinBlock(p[3])] + p[4] + p[6]


def p_join_block_0(p):
    """
    join_block : LKEY union_block RKEY rest_join_block
    """
    if p[4] != [] and isinstance(p[4][0], Filter):
        p[0] = [UnionBlock(p[2])] + p[4]
    if p[4] != [] and isinstance(p[4][0], Values):
        p[0] = [UnionBlock(p[2])] + p[4]
    elif p[4] != []:
        p[0] = [UnionBlock(p[2])] + [JoinBlock(p[4])]
    else:
        p[0] = [UnionBlock(p[2])]


def p_join_block_1(p):
    """
    join_block : bgp rest_join_block
    """
    p[0] = [p[1]] + p[2]


def p_rest_join_block_0(p):
    """
    rest_join_block : empty
    """
    p[0] = []


def p_rest_join_block_1(p):
    """
    rest_join_block : POINT bgp rest_join_block
    """
    p[0] = [p[2]] + p[3]


def p_rest_join_block_2(p):
    """
    rest_join_block : bgp rest_join_block
    """
    p[0] = [p[1]] + p[2]


def p_join_block_service_0(p):
    """
    join_block_service : LKEY bgp_service rest_join_block_service RKEY rest_join_block_service
    """
    jb_list = [p[2]] + p[3]
    if p[5] != [] and isinstance(p[5][0], Filter):
        p[0] = [UnionBlock([JoinBlock(jb_list)])] + p[5]
    elif isinstance(p[2], UnionBlock):
        p[0] = [p[2]] + p[3] + p[5]
    else:
        p[0] = [UnionBlock([JoinBlock(jb_list)])] + p[5]


def p_join_block_service_1(p):
    """
    join_block_service : bgp_service rest_join_block_service
    """
    p[0] = [p[1]] + p[2]


def p_rest_join_block_service_0(p):
    """
    rest_join_block_service : empty
    """
    p[0] = []


def p_rest_join_block_service_1(p):
    """
    rest_join_block_service : POINT bgp_service rest_join_block_service
    """
    p[0] = [p[2]] + p[3]


def p_rest_join_block_service_2(p):
    """
    rest_join_block_service : bgp_service rest_join_block_service
    """
    p[0] = [p[1]] + p[2]


def p_bgp_0(p):
    """
    bgp :  LKEY bgp UNION bgp rest_union_block RKEY
    """
    ggp = [JoinBlock([p[2]])] + [JoinBlock([p[4]])] + p[5]
    p[0] = UnionBlock(ggp)


def p_bgp_01(p):
    """
    bgp : bgp UNION bgp rest_union_block
    """
    ggp = [JoinBlock([p[1]])] + [JoinBlock([p[3]])] + p[4]
    p[0] = UnionBlock(ggp)


def p_bgp_1(p):
    """
    bgp : triple
    """
    p[0] = p[1]

# Filter ::=  'FILTER' Constraint (doesn't include FunctionCall yet)
def p_bgp_2(p):
    """
    bgp : FILTER LPAR expression RPAR
    """
    p[0] = Filter(p[3])


def p_bgp_3(p):
    """
    bgp : FILTER relational_expression
    """
    p[0] = Filter(p[2])


def p_bgp_4(p):
    """
    bgp : OPTIONAL LKEY group_graph_pattern RKEY
    """
    p[0] = Optional(p[3])


#def p_bgp_5(p):
#    """
#    bgp : LKEY join_block rest_union_block RKEY
#    """
#    bgp_arg = p[2] + p[3]
#    p[0] = UnionBlock(JoinBlock(bgp_arg))


def p_bgp_6(p):
    """
    bgp : LKEY join_block RKEY
    """
    if len(p[2]) == 1:
        p[0] = p[2][0]
    else:
        p[0] = JoinBlock(p[2])


def p_bgp_7(p):
    """
    bgp : VALUES LPAR var_list RPAR LKEY data_block RKEY
    """
    p[0] = Values(p[3], p[6])


def p_bgp_8(p):
    """
    bgp : VALUES var_list LKEY data_block RKEY
    """
    p[0] = Values(p[2], p[4])


def p_bgp_9(p):
    """
    bgp : BIND LPAR expression AS VARIABLE RPAR
    """
    p[0] = Bind(p[3], p[5])


def p_bgp_10(p):
    """
    bgp : service
    """
    p[0] = ([p[1]], [])


def p_bgp_11(p):
    """
    bgp : LKEY service RKEY
    """
    p[0] = ([p[2]], [])


def p_bgp_service_01(p):
    """
    bgp_service :  LKEY join_block_service UNION join_block_service rest_union_block_service RKEY
    """
    ggp = [JoinBlock(p[2])] + [JoinBlock(p[4])] + p[5]
    p[0] = UnionBlock(ggp)


def p_bgp_service_02(p):
    """
    bgp_service :  join_block_service UNION join_block_service rest_union_block_service
    """
    ggp = [JoinBlock(p[1])] + [JoinBlock(p[3])] + p[4]
    p[0] = UnionBlock(ggp)


def p_bgp_service_1(p):
    """
    bgp_service : triple
    """
    p[0] = p[1]


def p_bgp_service_2(p):
    """
    bgp_service : FILTER LPAR expression RPAR
    """
    p[0] = Filter(p[3])


def p_bgp_service_3(p):
    """
    bgp_service : FILTER relational_expression
    """
    p[0] = Filter(p[2])


def p_bgp_service_4(p):
    """
    bgp_service : OPTIONAL LKEY group_graph_pattern_service RKEY
    """
    p[0] = Optional(p[3])


def p_bgp_service_5(p):
    """
    bgp_service : LKEY join_block_service rest_union_block_service RKEY
    """
    bgp_arg = p[2] + p[3]
    p[0] = UnionBlock(JoinBlock(bgp_arg))


#############################################################################
# TODO: Test and adjust                                                     #
#############################################################################


def p_expression(p):
    """
    expression : conditional_or_expression
    """
    p[0] = p[1]


def p_conditional_or_expression(p):
    """
    conditional_or_expression : conditional_and_expression or_expr
    """
    if p[2] is not None:
        p[0] = Expression('||', p[1], p[2], 'logical')
    else:
        p[0] = p[1]


def p_or_expr_0(p):
    """
    or_expr : OR conditional_and_expression or_expr
    """
    if p[3] is not None:
        p[0] = Expression(p[1], p[2], p[3], 'logical')
    else:
        p[0] = p[2]


def p_or_expr_1(p):
    """
    or_expr : empty
    """
    p[0] = None


def p_conditional_and_expression(p):
    """
    conditional_and_expression : relational_expression and_expr
    """
    if p[2] is not None:
        p[0] = Expression('&&', p[1], p[2], 'logical')
    else:
        p[0] = p[1]


def p_and_expr_0(p):
    """
    and_expr : AND relational_expression and_expr
    """
    if p[3] is not None:
        p[0] = Expression(p[1], p[2], p[3], 'logical')
    else:
        p[0] = p[2]


def p_and_expr_1(p):
    """
    and_expr : empty
    """
    p[0] = None


def p_relational_expression_0(p):
    """
    relational_expression : numeric_expression EQUALS numeric_expression
    """
    p[0] = Expression(p[2], p[1], p[3], 'relational')


def p_relational_expression_1(p):
    """
    relational_expression : numeric_expression NEQUALS numeric_expression
    """
    p[0] = Expression(p[2], p[1], p[3], 'relational')


def p_relational_expression_2(p):
    """
    relational_expression : numeric_expression LESS numeric_expression
    """
    p[0] = Expression(p[2], p[1], p[3], 'relational')


def p_relational_expression_3(p):
    """
    relational_expression : numeric_expression GREATER numeric_expression
    """
    p[0] = Expression(p[2], p[1], p[3], 'relational')


def p_relational_expression_4(p):
    """
    relational_expression : numeric_expression LESSEQ numeric_expression
    """
    p[0] = Expression(p[2], p[1], p[3], 'relational')


def p_relational_expression_5(p):
    """
    relational_expression : numeric_expression GREATEREQ numeric_expression
    """
    p[0] = Expression(p[2], p[1], p[3], 'relational')


def p_relational_expression_6(p):
    """
    relational_expression : numeric_expression IN expression_list
    """
    p[0] = Expression(p[2], p[1], p[3], 'relational') 


def p_relational_expression_7(p):
    """
    relational_expression : numeric_expression NOT IN expression_list
    """
    inner_exp = Expression(p[3], p[1], p[4], 'relational')
    p[0] = Expression('!', inner_exp, exp_type = 'unary')


def p_relational_expression_8(p):
    """
    relational_expression : numeric_expression
    """
    p[0] = p[1]


def p_expression_list_0(p):
    """
    expression_list : LPAR expression RPAR
    """
    p[0] = [p[2]]


def p_expression_list_1(p):
    """
    expression_list : LPAR expression next_expr RPAR
    """
    p[0] = [p[2]] + p[3]


def p_next_expr_0(p):
    """
    next_expr : COMA expression next_expr
    """
    p[0] = [p[2]] + p[3]


def p_next_expr_1(p):
    """
    next_expr : empty
    """
    p[0] = []


def p_numeric_expression(p):
    """
    numeric_expression : additive_expression
    """
    p[0] = p[1]


def p_additive_expression(p):
    """
    additive_expression : multiplicative_expression arith_op
    """
    if p[2] != None:
        op, right = p[2]
        p[0] = Expression(op, p[1], right, 'arithmetic')
    else:
        p[0] = p[1]


def p_arith_op_0(p):
    """
    arith_op : plus_or_minus
    """
    p[0] = p[1]


def p_arith_op_1(p):
    """
    arith_op : empty
    """
    p[0] = None


def p_plus_or_minus(p):
    """
    plus_or_minus : PLUS multiplicative_expression arith_op
                   | MINUS multiplicative_expression arith_op
    """
    if p[3] != None:
        op, right = p[3]
        p[0] = p[1], Expression(op, p[2], right, 'arithmetic')
    else:
        p[0] = p[1], p[2]


def p_multiplicative_expression(p):
    """
    multiplicative_expression : unary_expression mult_div_unary
    """
    if p[2] is not None:
        op, right = p[2]
        p[0] = Expression(op, p[1], right, 'arithmetic')
    else:
        p[0] = p[1]


def p_mult_div_unary_0(p):
    """
    mult_div_unary : ALL unary_expression mult_div_unary
    """
    if p[3] != None:
        op, right = p[3]
        p[0] = p[1], Expression(op, p[2], right, 'arithmetic')
    else:
        p[0] = p[1], p[2]


def p_mult_div_unary_1(p):
    """
    mult_div_unary : DIV unary_expression mult_div_unary
    """
    if p[3] != None:
        op, right = p[3]
        p[0] = p[1], Expression(op, p[2], right, 'arithmetic')
    else:
        p[0] = p[1], p[2]


def p_mult_div_unary_2(p):
    """
    mult_div_unary : empty
    """
    p[0] = None


def p_unary_expression_0(p):
    """
    unary_expression : primary_expression
    """
    p[0] = p[1]


def p_unary_expression_1(p):
    """
    unary_expression : NEG primary_expression
                      | PLUS primary_expression
                      | MINUS primary_expression
    """
    p[0] = Expression(p[1], p[2], None, 'unary')


def p_primary_expression_0(p):
    """
    primary_expression : VARIABLE
    """
    p[0] = Argument(p[1], False)


def p_primary_expression_1(p):
    """
     primary_expression :  	bracketted_expression
                            | iri_or_function
                            | built_in_call
                            | rdf_literal
                            | numeric_literal
                            | boolean_literal
    """
    p[0] = p[1]


def p_bracketted_expression(p):
    """
    bracketted_expression : LPAR expression RPAR
    """
    p[0] = p[2]


# TODO: this is not implemented correctly yet, e.g. ID translation, etc. so the input must be URI
def p_iri_or_function_0(p):
    """
    iri_or_function : uri
    """
    p[0] = Argument(p[1], True)


# TODO: follow the spec. This is hardly simplified
def p_iri_or_function_1(p):
    """
    iri_or_function : type_casting VARIABLE
    """
    p[0] = Expression(p[1], Argument(p[2], False))


# TODO: follow the spec. This is hardly simplified
def p_iri_or_function_2(p):
    """
    iri_or_function : type_casting CONSTANT
    """
    p[0] = Expression(p[1], Argument(p[2], True))


# TODO: supposed to be uri and not constrained to the type_casting
def p_type_casting_0(p):
    """
    type_casting : DOUBLE 
                  | INTEGER
                  | DECIMAL
                  | FLOAT
                  | STRING
                  | BOOLEAN
                  | DATETIME
                  | NONPOSINT
                  | NEGATIVEINT
                  | LONG
                  | INT
                  | SHORT
                  | BYTE
                  | NONNEGINT
                  | UNSIGNEDLONG
                  | UNSIGNEDINT
                  | UNSIGNEDSHORT
                  | UNSIGNEDBYTE
                  | POSITIVEINT
    """
    p[0] = p[1]


# TODO: check whether the terminal from CONSTANT correct or not and test
def p_rdf_literal(p):
    """
    rdf_literal : CONSTANT
    """
    c = p[1].strip()
    if "@" in p[1]:
        p[0] = Argument(c[:c.find("^")], True, datatype="<" + xsd + "string>", lang=c[c.rfind("@")+1:], gen_type = str)
    elif xsd in p[1]:
        p[0] = Argument(c[:c.find("^")], True, datatype=c[c.rfind("^")+1:], gen_type = None)
    else:
        p[0] = Argument(p[1], True)


def p_numeric_literal_0(p):
   """
   numeric_literal : NUMBER
   """
   p[0] = Argument(p[1], True)


def p_numeric_literal_1(p):
   """
   numeric_literal : NUMBER POINT NUMBER
   """
   decimalNumber = str(p[1]) + p[2] + str(p[3])
   p[0] = Argument(decimalNumber, True)


# TODO: add double (with EXPONENT token)


def p_boolean_literal(p):
   """
   boolean_literal : TRUE
                    | FALSE
   """
   p[0] = Argument(p[1], True, datatype = xsd + 'boolean', gen_type = bool)


# TODO: builtInCall need to be rechecked, and developing new strategy, so that it won't take too much comparison during execution
def p_built_in_call_0(p):
    """
     built_in_call : STR LPAR expression RPAR
                    | LANG LPAR expression RPAR
                    | DATATYPE LPAR expression RPAR
                    | IRI LPAR expression RPAR
                    | URI LPAR expression RPAR
                    | ABS LPAR expression RPAR
                    | CEIL LPAR expression RPAR
                    | FLOOR LPAR expression RPAR
                    | ROUND LPAR expression RPAR
                    | STRLEN LPAR expression RPAR
                    | UCASE LPAR expression RPAR
                    | LCASE LPAR expression RPAR
                    | ENCODE_FOR_URI LPAR expression RPAR
                    | YEAR LPAR expression RPAR
                    | MONTH LPAR expression RPAR
                    | DAY LPAR expression RPAR
                    | HOURS LPAR expression RPAR
                    | MINUTES LPAR expression RPAR
                    | SECONDS LPAR expression RPAR
                    | TIMEZONE LPAR expression RPAR
                    | TZ LPAR expression RPAR
                    | MD5 LPAR expression RPAR
                    | SHA1 LPAR expression RPAR
                    | SHA256 LPAR expression RPAR
                    | SHA384 LPAR expression RPAR
                    | SHA512 LPAR expression RPAR
                    | ISIRI LPAR expression RPAR
                    | ISURI LPAR expression RPAR
                    | ISBLANK LPAR expression RPAR
                    | ISLITERAL LPAR expression RPAR
                    | ISNUMERIC LPAR expression RPAR
                    | BNODE LPAR expression RPAR
    """
    p[0] = Expression(p[1], p[3], exp_type = 'builtInUnary')


def p_built_in_call_1(p):
    """
     built_in_call : LANGMATCHES LPAR expression COMA expression RPAR
                    | CONTAINS LPAR expression COMA expression RPAR
                    | STRSTARTS LPAR expression COMA expression RPAR
                    | STRENDS LPAR expression COMA expression RPAR
                    | STRBEFORE LPAR expression COMA expression RPAR
                    | STRAFTER LPAR expression COMA expression RPAR
                    | STRLANG LPAR expression COMA expression RPAR
                    | STRDT LPAR expression COMA expression RPAR
                    | SAMETERM LPAR expression COMA expression RPAR
                    | REGEX LPAR expression COMA expression RPAR
                    | SUBSTR LPAR expression COMA expression RPAR
    """
    p[0] = Expression(p[1], p[3], p[5], 'builtInBinary')


def p_built_in_call_2(p):
    """
     built_in_call : RAND NIL
                    | NOW NIL
                    | UUID NIL
                    | STRUUID NIL
                    | BNODE NIL
    """
    p[0] = Expression(p[1], exp_type = 'builtInNil')


def p_built_in_call_3(p):
    """
     built_in_call : BOUND LPAR VARIABLE RPAR
    """
    p[0] = Expression(p[1], Argument(p[3], False), 'bound')


def p_built_in_call_4(p):
    """
     built_in_call : CONCAT expression_list
                    | COALESCE expression_list
    """
    p[0] = Expression(p[1], p[2], None)


def p_built_in_call_5(p):
    """
    built_in_call : REGEX LPAR expression COMA expression COMA expression RPAR
                      | IF LPAR expression COMA expression COMA expression RPAR
                      | SUBSTR LPAR expression COMA expression COMA expression RPAR
                      | REPLACE LPAR expression COMA expression COMA expression RPAR
    """
    p[0] = Expression(p[1], p[3], [p[5], p[7]])


def p_built_in_call_6(p):
    """
    built_in_call : REPLACE LPAR expression COMA expression COMA expression COMA expression RPAR
    """
    p[0] = Expression(p[1], p[3], [p[5], p[7], p[9]])


def p_built_in_call_7(p):
    """
    built_in_call : EXIST group_graph_pattern
    """
    p[0] = Expression(p[1], p[2])


def p_built_in_call_8(p):
    """
    built_in_call : NOT EXIST group_graph_pattern
    """
    op = p[1] + p[2]
    p[0] = Expression(op, p[3])


def p_built_in_call_9(p):
    """
    built_in_call : aggregate
    """
    p[0] = p[1]


#################################################################


def p_aggregate_0(p):
    """
    aggregate : COUNT LPAR distinct ALL RPAR
    """
    p[0] = Aggregate(Argument(p[4], True), p[3], p[1])


def p_aggregate_1(p):
    """
    aggregate : COUNT LPAR distinct expression RPAR
    """
    p[0] = Aggregate(p[4], p[3], p[1])


def p_aggregate_2(p):
    """
    aggregate : SUM LPAR distinct expression RPAR
    """
    p[0] = Aggregate(p[4], p[3], p[1])


def p_aggregate_3(p):
    """
    aggregate : AVG LPAR distinct expression RPAR
    """
    p[0] = Aggregate(p[4], p[3], p[1])


def p_aggregate_4(p):
    """
    aggregate : MIN LPAR distinct expression RPAR
    """
    p[0] = Aggregate(p[4], p[3], p[1])


def p_aggregate_5(p):
    """
    aggregate : MAX LPAR distinct expression RPAR
    """
    p[0] = Aggregate(p[4], p[3], p[1])


def p_aggregate_6(p):
    """
    aggregate : SAMPLE LPAR distinct expression RPAR
    """
    p[0] = Aggregate(p[4], p[3], p[1])


def p_aggregate_7(p):
    """
    aggregate : GROUP_CONCAT LPAR distinct expression RPAR
    """
    p[0] = Aggregate(p[4], p[3], p[1])


def p_aggregate_8(p):
    """
    aggregate : GROUP_CONCAT LPAR distinct expression SEMICOLON SEPARATOR EQUALS CONSTANT RPAR
    """
    p[0] = Aggregate(p[4], p[3], p[1], sep=p[8])


##################################################################


def p_triple_0(p):
    """
    triple : subject predicate object
    """
    p[0] = Triple(p[1], p[2], p[3])


def p_predicate_rdftype(p):
    """
    predicate : ID
    """
    if p[1] == 'a':
        value = '<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>'
        p[0] = Argument(value,True)
    else:
        print('raising')
        p_error(p[1])
        raise SyntaxError


def p_predicate_uri(p):
    """
    predicate : uri
    """
    p[0] = Argument(p[1], True)


def p_predicate_var(p):
    """
    predicate : VARIABLE
    """
    p[0] = Argument(p[1], False)


def p_subject_uri(p):
    """
    subject : uri
    """
    p[0] = Argument(p[1], True)


def p_subject_variable(p):
    """
    subject : VARIABLE
    """
    p[0] = Argument(p[1], False)


def p_object_uri(p):
    """
    object : uri
    """
    p[0] = Argument(p[1], True)


def p_object_variable(p):
    """
    object : VARIABLE
    """
    p[0] = Argument(p[1], False)


def p_object_constant(p):
    """
    object : CONSTANT
    """
    c = p[1].strip()
    p[0] = Argument(p[1], True)
    if xsd in p[1]:
        p[0] = Argument(c[:c.find("^")], True, datatype=c[c.rfind("^")+1:])
    if "@" in p[1]:
        p[0] = Argument(c[:c.find("^")], True, datatype="<" + xsd + "string>", lang=c[c.rfind("@")+1:])


def p_error(p):
    print(p)
    if isinstance(p, str):
        value = p
    else:
        value = p.value
    raise TypeError("unknown text at %r" % (value,))


parser = yacc.yacc(debug=0)


# Helpers
def parse(string):
    return parser.parse(urllib.parse.unquote(string), lexer=lexer)



"""
Created on Jul 10, 2011

Implements a Symmetric Hash Join (SHJ) operator.
The intermediate results are represented as queues.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Queue
from time import time
from DeTrusty.Operators.Join import Join
from .OperatorStructures import Table, Partition, Record


class SymmetricHashJoin(Join):

    def __init__(self, vars):
        self.left_table  = Table()
        self.right_table = Table()
        self.qresults    = Queue()
        self.vars        = vars

    def instantiate(self, d):
        newvars = self.vars - set(d.keys())
        return SymmetricHashJoin(newvars)

    def execute(self, left, right, out, processqueue=Queue()):
        # Executes the Symmetric Hash Join.
        self.left     = left
        self.right    = right
        self.qresults = out

        # Initialize tuples.
        tuple1 = None
        tuple2 = None

        # Get the tuples from the queues.
        while (not(tuple1 == "EOF") or not(tuple2 == "EOF")):

            # Try to get and process tuple from left queue.
            if not(tuple1 == "EOF"):
                try:
                    tuple1 = self.left.get(False)
                    #print "Tuples in right table", len(self.right_table.partitions[0].records)
                    self.insertAndProbe(tuple1, self.left, self.left_table, self.right_table)
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.left.get(False), when the queue is empty.
                    # TypeError: in att = att + tuple[var], when the tuple is "EOF".
                    pass

            # Try to get and process tuple from right queue.
            if not(tuple2 == "EOF"):
                try:
                    tuple2 = self.right.get(False)
                    self.insertAndProbe(tuple2, self.right, self.right_table, self.left_table)
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    # TypeError: in att = att + tuple[var], when the tuple is "EOF".
                    pass

#            # Process tuples.
#            if (not(tuple1 == None) and not(tuple1 == "EOF")):
#                #print "Tuples in right table", len(self.right_table.partitions[0].records)
#                self.insertAndProbe(tuple1, self.left, self.left_table, self.right_table)
#            if (not(tuple2 == None) and not(tuple2 == "EOF")):
#                self.insertAndProbe(tuple2, self.right, self.right_table, self.left_table)
#

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return

    def insertAndProbe(self, tuple, input, table1, table2):
        # Insert the tuple in its corresponding partition and probe.

        # Get the attribute(s) to apply hash.
        att = ''
        for var in self.vars:
            att = att + tuple[var]
        i = hash(att) % table1.size;

        # Insert record in partition.
        record = Record(tuple, time(), 0)
        table1.insertRecord(i, record)

        # Probe the record against its partition in the other table.
        self.probe(record, table2.partitions[i], self.vars)

    def probe(self, record, partition, var):
        # Probe a tuple if the partition is not empty.
        #count = 0
        if partition:

            # For every record in the partition, check if it is duplicated.
            # Then, check if the tuple matches for every join variable.
            # If there is a join, concatenate the tuples and produce result.

            for r in partition.records:
                #count = count + 1
                if self.isDuplicated(record, r):
                    break

                for v in var:
                    join = True
                    if record.tuple[v] != r.tuple[v]:
                        join = False
                        break

                if join:
                    res = record.tuple.copy()
                    res.update(r.tuple)
                    self.qresults.put(res)
                    #print count, res

    def isDuplicated(self, record1, record2):
        # Verify if the tuples has been already probed.
        return not record1.ats >= record2.ats



"""
NestedHashOptionalFilter.py

Implements a depending operator, similar to block nested join and symmetric hash join

Autor: Maribel Acosta
Autor: Maria-Esther Vidal
Autor: Gabriela Montoya
Date: January 29th, 2014
"""

from multiprocessing import Queue, Process
from time import time
import string, sys
from queue import Empty
from DeTrusty.Operators.Optional import Optional
from .OperatorStructures import Table, Partition, Record

WINDOW_SIZE = 10


class NestedHashOptionalFilter(Optional):

    def __init__(self, vars_left, vars_right):
        self.left_table = dict()
        self.right_table = dict()
        self.qresults = Queue()
        self.bag = []
        self.vars_left = set(vars_left)
        self.vars_right = set(vars_right)
        self.vars = list(self.vars_left & self.vars_right)

    def instantiate(self, d):
        newvars_left = self.vars_left - set(d.keys())
        newvars_right = self.vars_right - set(d.keys())
        return NestedHashOptionalFilter(newvars_left, newvars_right)

    def instantiateFilter(self, d, filter_str):
        newvars_left = self.vars_left - set(d)
        newvars_right = self.vars_right - set(d)
        return NestedHashOptionalFilter(newvars_left, newvars_right)

    def execute(self, left_queue, right_operator, out, processqueue=Queue()):
        # print "execute NestedHashOptionalFilter"
        self.left_queue = left_queue
        self.right_operator = right_operator
        self.qresults = out

        tuple1 = None
        tuple2 = None
        right_queues = dict()
        filter_bag = []
        count = 0
        while (not (tuple1 == "EOF") or (len(right_queues) > 0)):

            try:
                tuple1 = self.left_queue.get(False)
                # print "tuple1", tuple1
                # Try to get and process tuple from left queue
                if not (tuple1 == "EOF"):
                    # tuple1 = self.left_queue.get(False)
                    # print "tuple1: "+str(tuple1)
                    self.bag.append(tuple1)
                    instance = self.probeAndInsert1(tuple1, self.right_table, self.left_table, time())
                    # print "sali de probe and insert 1 con tuple", tuple1
                    if instance:  # the join variables have not been used to
                        # instanciate the right_operator
                        filter_bag.append(tuple1)
                    # print "filter_bag", len(filter_bag)

                    if len(filter_bag) >= WINDOW_SIZE:
                        new_right_operator = self.makeInstantiation(filter_bag, self.right_operator)
                        # print "Here in makeInstantation with filter"
                        # resource = self.getResource(tuple1)
                        queue = Queue()
                        right_queues[count] = queue
                        new_right_operator.execute(queue)
                        filter_bag = []
                        count = count + 1

                else:
                    if (len(filter_bag) > 0):
                        # print "here", len(filter_bag), filter_bag
                        new_right_operator = self.makeInstantiation(filter_bag, self.right_operator)
                        # resource = self.getResource(tuple1)
                        queue = Queue()
                        right_queues[count] = queue
                        new_right_operator.execute(queue)
                        filter_bag = []
                        count = count + 1

            except Empty:
                pass
            except Exception as e:
                # print "Unexpected error:", sys.exc_info()[0]
                # print e
                pass

            toRemove = []  # stores the queues that have already received all its tuples
            # print "right_queues", right_queues
            for r in right_queues:
                try:
                    q = right_queues[r]
                    tuple2 = None
                    while (tuple2 != "EOF"):
                        tuple2 = q.get(False)

                        if (tuple2 == "EOF"):
                            toRemove.append(r)
                        else:
                            resource = self.getResource(tuple2)
                            for v in self.vars:
                                del tuple2[v]
                            # print "new tuple2", tuple2
                            self.probeAndInsert2(resource, tuple2, self.left_table, self.right_table, time())
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    # TypeError: in att = att + tuple[var], when the tuple is "EOF".
                    # print "Unexpected error:", sys.exc_info()
                    pass

            for r in toRemove:
                del right_queues[r]

        # This is the optional: Produce tuples that haven't matched already.
        for tuple in self.bag:
            res_right = {}
            for var in self.vars_right:
                res_right.update({var: ''})
            res = res_right
            res.update(tuple)
            self.qresults.put(res)

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return

    def getResource(self, tuple):
        resource = ''
        for var in self.vars:
            val = tuple[var]
            if "^^<" in val:
                val = val[:val.find('^^<')]
            resource = resource + val
        return resource

    def makeInstantiation(self, filter_bag, operator):
        filter_str = ''
        filter_str = " . ".join(map(str, operator.tree.service.filters))
        new_vars = ['?' + v for v in self.vars]  # TODO: this might be $
        # print "operator type ",operator
        # print "making instantiation join filter", filter_bag
        # When join variables are more than one: FILTER ( )
        if len(self.vars) >= 1:
            filter_str += ' . FILTER (__expr__)'

            or_expr = []
            for tuple in filter_bag:
                and_expr = []
                for var in self.vars:
                    # aux = "?" + var + "==" + tuple[var]
                    v = tuple[var]
                    if v.find("http") == 0:  # uris must be passed between < .. >
                        v = "<" + v + ">"
                    else:
                        v = '"' + v + '"'
                    v = "?" + var + "=" + v
                    and_expr.append(v)

                or_expr.append('(' + ' && '.join(and_expr) + ')')
            filter_str = filter_str.replace('__expr__', ' || '.join(or_expr))

        # When join variables is a singleton: FILTER ?var IN ()
        # else:
        #
        #    filter_str = ' . FILTER ?' + list(self.vars)[0] + ' IN (__expr__)'
        #
        #    expr = []
        #    for tuple in filter_bag:
        #        v = tuple[list(self.vars)[0]]
        #        if string.find(v, "http") == 0: # uris must be passed between < .. >
        #            v = "<"+v+">"
        #        expr.append(v)
        #    filter_str = filter_str.replace('__expr__', ' , '.join(expr))

        new_operator = operator.instantiateFilter(set(new_vars), filter_str)
        # print "type(new_operator)", type(new_operator)
        return new_operator

    def probeAndInsert1(self, tuple, table1, table2, time):
        # print "in probeAndInsert1", tuple
        record = Record(tuple, time, 0)
        r = self.getResource(tuple)
        # print "resource", r, tuple
        if r in table1:
            records = table1[r]
            for t in records:
                if t.ats > record.ats:
                    continue
                x = t.tuple.copy()
                x.update(tuple)
                self.qresults.put(x)
                # Delete tuple from bag.
                try:
                    self.bag.remove(tuple)
                except ValueError:
                    pass

        p = table2.get(r, [])
        i = (p == [])
        p.append(record)
        table2[r] = p
        return i

    def probeAndInsert2(self, resource, tuple, table1, table2, time):
        # print "probeAndInsert2", resource, tuple
        record = Record(tuple, time, 0)
        if resource in table1:
            records = table1[resource]
            for t in records:
                if t.ats > record.ats:
                    continue
                x = t.tuple.copy()
                x.update(tuple)
                self.qresults.put(x)
                # Delete tuple from bag.
                try:
                    self.bag.remove(t.tuple)
                except ValueError:
                    pass

        p = table2.get(resource, [])
        p.append(record)
        table2[resource] = p



"""
Created on Jul 10, 2011

Implements the Xoffset operator.
The intermediate results are represented in a queue.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Queue


class Xoffset(object):

    def __init__(self, vars, offset):
        self.input = Queue()
        self.qresults = Queue()
        self.vars = vars
        self.offset = int(offset)

    def execute(self, left, dummy, out, processqueue=Queue()):
        # Executes the Xoffset.
        self.left = left
        self.qresults = out
        tuple = self.left.get(True)
        count = 0

        # OFFSET
        while ((count < self.offset) and (tuple != "EOF")):
            count = count + 1
            tuple = self.left.get(True)

        # Producing the remaining results.
        while (tuple != "EOF"):
            self.qresults.put(tuple)
            tuple = self.left.get(True)

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return



"""
Created on Jul 10, 2011

Implements the Xproject operator.
The intermediate results are represented in a queue.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Queue
from . import Xexpression
from DeTrusty.Sparql.Parser.services import Expression, Aggregate


class Xproject(object):

    def __init__(self, vars):
        self.input = Queue()
        self.qresults = Queue()
        self.vars = vars

    def execute(self, left, dummy, out, processqueue=Queue()):
        # Executes the Xproject.
        self.left = left
        self.qresults = out
        tuple = self.left.get(True)
        while not (tuple == "EOF"):
            res = {}
            if len(self.vars) == 0:
                self.qresults.put(dict(tuple))
            else:
                for var in self.vars:
                    alias = None
                    if isinstance(var, Expression) or isinstance(var, Aggregate):
                        tmp = Xexpression.simplifyExp(var, tuple)
                        tuple.update({var.alias[1:]: str(tmp)})
                        var = var.alias[1:]
                    else:
                        if var.alias is not None:
                            alias = var.alias[1:]
                        var = var.name[1:]
                    aux = tuple.get(var, '')
                    if alias is not None:
                        res.update({alias: aux})
                    else:
                        res.update({var: aux})
                self.qresults.put(res)

            tuple = self.left.get(True)

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return



"""
Created on Jul 10, 2011

Implements the Xunion operator.
The intermediate results are represented in a queue.

@author: Maribel Acosta Deibe

Modified 2022-10-17 by Philipp D. Rohde to speed up merging both input queues.
"""

from multiprocessing import Queue, Process
from queue import Empty
from DeTrusty.Operators.Union import _Union


class Xunion(_Union):

    def __init__(self, vars_left, vars_right):
        self.left = Queue()
        self.right = Queue()
        self.qresults = Queue()
        self.vars_left = vars_left
        self.vars_right = vars_right

    def instantiate(self, d):
        newvars_left = self.vars_left - set(d.keys())
        newvars_right = self.vars_right - set(d.keys())
        return Xunion(newvars_left, newvars_right)

    def instantiateFilter(self, instantiated_vars, filter_str):
        newvars_left = self.vars_left - set(instantiated_vars)
        newvars_right = self.vars_right - set(instantiated_vars)
        return Xunion(newvars_left, newvars_right)

    def execute(self, left, right, out, processqueue=Queue()):
        # Executes the Xunion.
        self.left = left
        self.right = right
        self.qresults = out
        # print "left", hex(id(left)), "right", hex(id(right)), "out", hex(id(out))

        # Identify the kind of union to perform.
        if self.vars_left == self.vars_right:
            self.same_variables()
        else:
            self.different_variables()

        # Put EOF in queue and exit.
        self.qresults.put("EOF")

    @staticmethod
    def __insert_result(in_, out, vars=None):
        # Puts the results from one input queue to the output queue.
        try:
            tuple_ = in_.get()
            while tuple_ != 'EOF':
                if vars is not None:
                    res = {}
                    res.update(vars)
                    res.update(tuple_)
                else:
                    res = tuple_
                out.put(res)
                tuple_ = in_.get()
        except Empty:
            pass

    def same_variables(self):
        # Executes the Xunion operator when the variables are the same.
        p_left = Process(target=self.__insert_result, args=(self.left, self.qresults))
        p_left.start()

        p_right = Process(target=self.__insert_result, args=(self.right, self.qresults))
        p_right.start()

        p_left.join()
        p_right.join()

    def different_variables(self):
        # Executes the Xunion operator when the variables are not the same.
        # Initialize empty tuples.
        v1 = {}
        v2 = {}

        # Add empty values to variables of the other argument.
        for v in self.vars_right:
            v1.update({v: ''})

        for v in self.vars_left:
            v2.update({v: ''})

        p_left = Process(target=self.__insert_result, args=(self.left, self.qresults, v1))
        p_left.start()

        p_right = Process(target=self.__insert_result, args=(self.right, self.qresults, v2))
        p_right.start()

        p_left.join()
        p_right.join()



__author__ = "Philipp D. Rohde"

import logging


def get_logger(name, filename=None, level=logging.INFO, file_and_console=False):
    logger = logging.getLogger(name)
    logger.setLevel(level)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    if filename is not None:
        file_handler = logging.FileHandler(filename)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    if filename is None or file_and_console:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    return logger



"""
Created on Jul 10, 2011

Implements a Nested Loop Optional operator.
The intermediate results are represented as lists.

@author: Maribel Acosta Deibe
"""

from time import time
from multiprocessing import Queue
from DeTrusty.Operators.Optional import Optional
from .OperatorStructures import Table, Partition, Record


class NestedLoopOptional(Optional):

    def __init__(self, vars_left, vars_right):
        self.left_table  = Table()
        self.right_table = Table()
        self.results     = []
        self.vars_left   = set(vars_left)
        self.vars_right  = set(vars_right)
        self.vars        = list(self.vars_left & self.vars_right)

    def instantiate(self, d):
        newvars_left = self.vars_left - set(d.keys())
        newvars_right = self.vars_right - set(d.keys())
        return NestedLoopOptional(newvars_left, newvars_right)

    def execute(self, qleft, qright, out, processqueue=Queue()):
        # Executes the Nested Loop Optional.
        self.left = []
        self.right = qright
        self.qresults = out

        # Initialize tuple.
        tuple = None

        # GEt tuple from queue.
        while not(tuple == "EOF"):
            tuple = qleft.get(True)
            self.left.append(tuple)

        # Get the variables to join.
        if (len(self.left)>1):
            # Iterate over the lists to get the tuples.
            while (len(self.left)>1):
                tuple = self.left.pop(0)
                self.insertAndProbe(tuple)

        # Put all the results in the output queue.
        while self.results:
            self.qresults.put(self.results.pop(0))

        # Put EOF in queue and exit.
        self.qresults.put("EOF")

    def insertAndProbe(self, tuple):
        # Executes the Nested Loop Join.

        # Get the attribute(s) to apply hash.
        att1 = ''
        for var in self.vars:
            att1 = att1 + tuple[var]
        i = hash(att1) % self.left_table.size

        # Create record (tuple, ats, dts).
        record = Record(tuple, time(), 0)

        # Insert record in its corresponding partition.
        self.left_table.insertRecord(i, record)

        # Probe the record against its partition in the other table.
        self.probe(record, i, self.right_table.partitions[i], self.vars, self.right)

    def probe(self, record, i, partition, var, right):
        # Probe a tuple if the partition is not empty.
        if partition:
            anyjoin = False

            # For every record in the partition, check if it is duplicated.
            # Then, check if the tuple matches for every join variable.
            # If there is a join, concatenate the tuples and produce result.
            # If the partition was empty, or any join was produced, then contact the source.
            # If no match in the contacted source were produced, then concatenate with empty tuple.
            for r in partition.records:

                if self.isDuplicated(record, r):
                    break

                for v in var:
                    join = True
                    if record.tuple[v] != r.tuple[v]:
                        join = False
                        break

                if join:
                    anyjoin = True
                    res = record.tuple.copy()
                    res.update(r.tuple)
                    self.results.append(res)

            # Empty partition or no matches were found.
            if ((len(partition.records) == 0) or not(anyjoin)):
                instances = []
                for v in var:
                    instances = instances + [record.tuple[v]]

                # Contact the source.
                qright = Queue()
                right.execute(self.vars, instances, qright)

                # This is the join.
                # Insert in right table, and produce the results.
                # Get the tuples from right queue.
                rtuple = qright.get(True)
                if (not(rtuple == "EOF")):
                    while (not(rtuple == "EOF")):
                        res2 = rtuple.copy()

                        for v in var:
                            res2.update({v:record.tuple[v]})

                        reg = Record(res2, time(), 0)
                        self.right_table.insertRecord(i, reg)

                        res = rtuple.copy()
                        res.update(record.tuple)
                        self.results.append(res)

                        rtuple = qright.get(True)

                # This is the optional.
                # Construct empty tuple, insert in table, and produce the results.
                else:
                    res = {}

                    for k in self.right.atts:
                        res.update({k:''})

                    reg = Record(res, time(), 0)
                    self.right_table.insertRecord(i, reg)

                    res.update(record.tuple)
                    self.results.append(res)

    def isDuplicated(self, record1, record2):
        # Verify if the tuples has been already probed.
        return not record1.ats >= record2.ats



"""
Created on Dec 11, 2013

Implements the Xdistinct operator.
The intermediate results are represented in a queue.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Queue


class Xdistinct(object):

    def __init__(self, vars):
        # self.input       = Queue()
        self.qresults = Queue()
        self.vars = vars
        self.bag = {}

    def execute(self, left, dummy, out, processqueue=Queue()):
        # Executes the Xdistinct.
        self.left = left
        self.qresults = out
        tuple = self.left.get(True)

        while (not (tuple == "EOF")):
            str_tuple = str(sorted(tuple.items()))
            get = self.bag.get(str_tuple, False)

            if not (get):
                self.qresults.put(tuple)
                self.bag.update({str_tuple: True})
            tuple = self.left.get(True)

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return



"""
Created on 30/08/2021.

Implements the Xgroupby operator.
The intermediate results are represented in a queue.

@author: Avellino
"""

# TODO: the grouping is quite wrong in case of implicit grouping, reference: 18.2.4.1 in spec.
from multiprocessing import Queue

class Xgroupby(object):                                                    # ensure backward comp.

    def __init__(self, args):
        self.qresults = Queue()                                            # end res.
        self.args = args                                                   # GROUP BY vars

    def execute(self, left, dummy, out, processqueue=Queue()):
        self.left = left
        self.qresults = out
        arg_list = list()
        tmp = list()                                                       # [tup1,tup2,...]

        for arg in self.args:
            arg_list.append(arg.name[1:])

        tuple = self.left.get(True)                                     
        # print(tuple)

        while (tuple != "EOF"):
            match = [0] * len(tmp)
            append = True

            for i in range(len(tmp)):
                tmp_dict = tmp[i]
                for arg in self.args:                                       # type(arg) always Argument
                    if arg.alias is not None:
                        if tuple[arg.name[1:]] == tmp_dict[arg.alias[1:]]:
                            match[i] += 1                                   # record matches based on tmp index
                    else:
                        if tuple[arg.name[1:]] == tmp_dict[arg.name[1:]]:
                            match[i] += 1

            for j in range(len(match)):
                if match[j] == len(self.args): 
                    append = False
                    for var in tuple:
                        if var not in arg_list:
                            tmp_dict = tmp[j]
                            tmp_dict[var].append(tuple[var])

            if append:
                for arg in self.args:
                    if arg.alias is not None:
                        tuple.update({arg.alias[1:]: tuple[arg.name[1:]]})
                for key in tuple:
                    if key not in arg_list:
                        tuple.update({key: [tuple[key]]})
                tmp.append(tuple)

            tuple = self.left.get(True)

        ###DEBUG###
        # print("GROUP BY vars:", arg_list)
        # for i in range(len(tmp)):
            # tmp_dict = tmp[i]
            # for attr in tmp_dict:
                # string = attr + ": " 
                # if type(tmp_dict[attr]) == list:
                    # string = string + str(len(tmp_dict[attr]))
                # else:
                    # string = string + tmp_dict[attr]
                # print(string)

        tmp.reverse()
        while tmp:
            self.qresults.put(tmp.pop())
        self.qresults.put("EOF")
        return
